function [gv,gvb,bl,dmvol,fedpars, grain, blobondet] = gtFedTestLoadData_v2(fedpars, grain, parameters, dmvol)
% [gv,gvb,bl,dmvol,fedpars] = gtFedTestLoadData_v2(fedpars, grain, parameters, dmvol)
%
% Sets up the framework, variables, random strain distribution for simulated data.
%  In the grain volume everything is defined in voxels, no dimensioning with pixel size.
%  In the projections and diff. blobs, everything is defined in detector pixels (here comes the pixel size and setup geometry into play).
%    Thus the dimensioning/scaling happens in the forward projection
%    functions and their derivative functions, since they relate the
%    grain voxels to the detector pixels.
%
% INPUT
%   fedpars    - FED parameters
%                Some important ones:
%                  grainsize   - no. of voxels in grain
%                  grainenv    = grainsize
%                  ngv         = grainsize
%                  gvsize      = [1 1 1]
%                  loadbl      - blobs ID-s in grain.allblobs to be generated
%                  dcomps      - for rotations only use [1 1 1 0 0 0 0 0 0]
%                  blobsizeadd - blob volume padding
%                  vox000sam   - (X,Y,Z) coordinates of voxel (0,0,0) in the Sample reference frame
%   grain      - grain data as output from Indexter
%   parameters - parameters as in DCT paramaters file
%   dmvol      - all deformation components per voxel for the entire volume
%                (nx,ny,nz,ndef); if empty, random distribution is applied
%                with max limits as in fedpars.dapplim and max gradients
%                as in fedpars.dappgradlim

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Load and update geometry parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

labgeo = parameters.labgeo;
samgeo = parameters.samgeo;

labgeo.omstep = gtGetOmegaStepDeg(parameters);
labgeo.detnorm  = gtMathsCross(labgeo.detdiru, labgeo.detdirv); % unit vector

% Detector position
%   originally in detector pixels !!!
%   change to grain voxels !!!
%   pixel sizes in mm-s !!!
labgeo.detrefuv  = [labgeo.detrefu; labgeo.detrefv];

% Rotation matrix components:
labgeo.rotcomp = gtMathsRotationMatrixComp(labgeo.rotdir', 'col');

% mm/pixelsize
labgeo.detscaleu = 1 / parameters.labgeo.pixelsizeu ;
labgeo.detscalev = 1 / parameters.labgeo.pixelsizev ;

% Rotation tensor components
labgeo.Qdet = gtFedDetectorProjectionTensor( ...
    labgeo.detdiru, labgeo.detdirv, labgeo.detscaleu, labgeo.detscalev);

% Update parameters
parameters.labgeo = labgeo;

% Compute detector origin from the reference position
% detorig = labgeo.detrefpos - ...
%             labgeo.detrefu*labgeo.detdiru*labgeo.pixelsizeu - ...
%             labgeo.detrefv*labgeo.detdirv*labgeo.pixelsizev;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Load grain parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

grsize = fedpars.grainsize;      % x,y,z number of voxels in grain
gvsize = fedpars.gvsize;         % x,y,z number of voxels in an element

% Grain volume
grvol  = gtFedTestGenerateGrainVolume(grsize);
fedpars.grainvol = grvol;

% Determine number of volume elements along X,Y,Z, based on grain volume
% and element size:
ngv(1) = ceil(size(grvol,1)/gvsize(1));
ngv(2) = ceil(size(grvol,2)/gvsize(2));
ngv(3) = ceil(size(grvol,3)/gvsize(3));

fedpars.ngv = ngv;

% Size of grain envelope
grenv(1) = ngv(1)*gvsize(1);
grenv(2) = ngv(2)*gvsize(2);
grenv(3) = ngv(3)*gvsize(3);

fedpars.grainenv = grenv;

% Padded grain volume
grvol_pad = false(grenv);
grvol_pad(1:size(grvol,1),1:size(grvol,2),1:size(grvol,3)) = grvol;
grvol     = grvol_pad;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Volume sampling and volume elements
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Allocating memory for elements ...')
tic

nv    = ngv(1)*ngv(2)*ngv(3);          % total number of vol elements
svvol = gvsize(1)*gvsize(2)*gvsize(3); % size of sampling volume

nd = sum(fedpars.dcomps);

% Element XYZ index
[gv.ind(:, 1), gv.ind(:, 2), gv.ind(:, 3)] = ind2sub(ngv, 1:nv);
gv.ind = gv.ind';

% Elements' positions in the sample volume in mm!
gv.cs = fedpars.vox000sam(ones(nv,1), :)' + gv.ind .* parameters.recgeo.voxsize(ones(nv,1), :)';
%gv.cs1 = gv.cs;

% Grain center position in sample in mm!
%gc = mean(double(gv.cs),2);
gc = fedpars.vox000sam + (grsize/2 + [0.5, 0.5, 0.5]) .* parameters.recgeo.voxsize;
fedpars.graincentsam = gc;
grain.center = gc;

% Element diffracting power
gv.pow = ones(1, nv);

% Element actual strain state
gv.d = zeros(nd, nv);

% Element change in strain state
gv.dd = zeros(nd, nv);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Computing allblobs info
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

grain = gtCalculateGrain(grain, parameters);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Iinitial blob parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Generating blob data...')
tic

% Blobs to load and use
lbl = fedpars.loadbl;

% Number of blobs to consider
nbl = length(fedpars.loadbl(:));

% Create a blank blob structure
tmp       = gtFedBlobDefinition();
bl(1:nbl) = tmp;

% Loop through blobs
blobsdiffaway = [];
blobondet     = [];

for ii = 1:nbl
    % ID-s
    bl(ii).blobIDfed   = ii;
    bl(ii).blobIDgrain = fedpars.loadbl(ii);

    % Load diffraction paramaters of blobs

    % Plane families
    bl(ii).hkl       = grain.allblobs.hkl(lbl(ii), :);
    bl(ii).hklsp     = grain.allblobs.hklsp(lbl(ii), :);
    bl(ii).thetatype = grain.allblobs.thetatype(lbl(ii));

    % Initial strained signed plane normal in SAMPLE reference
    bl(ii).plsam = grain.allblobs.pl(lbl(ii), :);

    % Initial strained non-signed plane normal in SAMPLE reference
    bl(ii).plorig = grain.allblobs.plorig(lbl(ii), :);

    % Initial strained signed plane normal in the diffraction position in
    % LAB reference
    bl(ii).pl0 = grain.allblobs.pllab(lbl(ii), :);

    % Initial strained diffraction angles and rotation matrix
    bl(ii).th0    = grain.allblobs.theta(lbl(ii));     % initial theta
    bl(ii).sinth0 = grain.allblobs.sintheta(lbl(ii));  % initial sin(theta)
    bl(ii).eta0   = grain.allblobs.eta(lbl(ii));       % initial eta
    bl(ii).om0    = grain.allblobs.omega(lbl(ii));     % initial omega
    bl(ii).omind  = grain.allblobs.omind(lbl(ii));     % initial omega index (1..4)
    bl(ii).S0     = grain.allblobs.srot(:, :, lbl(ii)); % initial rotation tensor

    % Initial strained diffraction vector in LAB reference
    bl(ii).dvec0 = grain.allblobs.dvec(lbl(ii), :);

    % Initial strained diffraction vector in SAMPLE reference
    bl(ii).dvecsam = grain.allblobs.dvecsam(lbl(ii), :);

    % Initial centroid in image
    bl(ii).u0im = grain.allblobs.uvw(lbl(ii), :);

    if isempty(bl(ii).u0im)
        blobsdiffaway = [blobsdiffaway, ii];
    else
        % Padding around blob volume
        bl(ii).addbl = fedpars.blobsizeadd;

        % Check if blob hits the detector area
        uok = (bl(ii).u0im(1) < parameters.acq.xdet) & (bl(ii).u0im(1) > 1);
        vok = (bl(ii).u0im(2) < parameters.acq.ydet) & (bl(ii).u0im(2) > 1);

        if (uok && vok)
            blobondet = [blobondet, ii];
        end
    end
end

disp('Blobs inside the detector area are:')
disp(blobondet)

if ~isempty(blobsdiffaway)
    disp('The following projections (blobs) are directed away from the detector plane.')
    disp(blobsdiffaway)
    error('Remove projections (blobs) that do not hit the detector.')
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Arrows of projection vectors
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Quiver figure of projection vectors
dvecsam = vertcat(bl(:).dvecsam);

figure('name','Projection vectors')
quiver3(zeros(nbl,1), zeros(nbl,1), zeros(nbl,1), ...
        dvecsam(:,1), dvecsam(:,2), dvecsam(:,3));
axis equal
hold on
for ii = 1:length(bl)
    text(dvecsam(ii,1), dvecsam(ii,2), dvecsam(ii,3), ...
        num2str(ii), 'VerticalAlignment', 'Bottom', 'Color','m')
    text(dvecsam(ii,1), dvecsam(ii,2), dvecsam(ii,3), ...
        num2str(bl(ii).blobIDgrain), 'VerticalAlignment','Top','Color','r')
end
title('blob ID in FED (magenta), blob ID in grain (red)')
drawnow();

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Preallocate blob volumes
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

maxdiag = ceil(sqrt(grenv*grenv'));

for ii = 1:nbl
    % Bounding box sizes and boundaries for processing
    bl(ii).bbsize = [maxdiag + 2*bl(ii).addbl(1), ...
                     maxdiag + 2*bl(ii).addbl(2), ...
                     2 + 2*bl(ii).addbl(3) ];

    tmp          = round( bl(ii).u0im(1) - bl(ii).bbsize(1)/2 );
    bl(ii).bbuim = [tmp, tmp + bl(ii).bbsize(1) - 1];

    tmp          = round( bl(ii).u0im(2) - bl(ii).bbsize(2)/2 );
    bl(ii).bbvim = [tmp, tmp + bl(ii).bbsize(2) - 1];

    tmp          = round( bl(ii).u0im(3) - bl(ii).bbsize(3)/2 );
    bl(ii).bbwim = [tmp, tmp + bl(ii).bbsize(3) - 1];

    % Blob origin and center [u,v,w]:
    bl(ii).bbor   = [bl(ii).bbuim(1)-1, bl(ii).bbvim(1)-1, bl(ii).bbwim(1)-1];

    bl(ii).bbcent = [round( (bl(ii).bbuim(1) + bl(ii).bbuim(2))/2 ),...
                     round( (bl(ii).bbvim(1) + bl(ii).bbvim(2))/2 ),...
                     round( (bl(ii).bbwim(1) + bl(ii).bbwim(2))/2 ) ];

    % Projection parameters
    bl(ii).proj_geom = gtGeoProjForReconstruction( ...
        bl(ii).dvecsam, bl(ii).om0, gc, ...
        [bl(ii).bbuim(1) bl(ii).bbvim(1) bl(ii).bbsize(1:2)], ...
        [], parameters.labgeo, parameters.samgeo, parameters.recgeo, ...
        'ASTRA_grain');

    %%% Needed for FED
%     bl(ii).int  = zeros(bl(ii).bbsize, fedpars.bltype);
%     bl(ii).intd = zeros(bl(ii).bbsize, fedpars.bltype);
    bl(ii).intm = zeros(bl(ii).bbsize, fedpars.bltype);
    bl(ii).mask = ones(bl(ii).bbsize, fedpars.bltype);

    % u in blob = u in image minus u at blob origin
    bl(ii).u0bl = bl(ii).u0im - bl(ii).bbor;
end

toc

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Initial element parameters
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Generating element data ...')
tic

gvcs = gv.cs;

for ii = nbl:-1:1
    gvb(ii).uc0bl   = zeros(size(gvcs));
    gvb(ii).uc      = zeros(size(gvcs));
    gvb(ii).ucd     = zeros(size(gvcs));
    gvb(ii).ucblmod = zeros(size(gvcs));
    gvb(ii).ucbl8   = zeros(1, size(gvcs,2)*8);
    gvb(ii).ucbl8ini= zeros(1, size(gvcs,2)*8);
    gvb(ii).ucbl8ok = 0;

    % Expand rotation tensors
    subs = {(1:3)', (1:3)', ones(1,nv)};
    rot_l2s  = bl(ii).S0(subs{:});

    % Expand diffraction vector
    dvec = bl(ii).dvec0';
    dvec = dvec(:,ones(1,nv));

    % Expand omega
    om = bl(ii).om0(1,ones(1,nv));

    rot_s2l = permute(rot_l2s, [2 1 3]);
    gvcs_lab = gtGeoSam2Lab(gvcs', rot_s2l, labgeo, samgeo, false);

    % Initial absolut u of gv center
    uc0im = gtFedPredictUVWMultiple([], dvec, gvcs_lab', ...
        labgeo.detrefpos', labgeo.detnorm', labgeo.Qdet, ...
        labgeo.detrefuv, om, labgeo.omstep );

    % Initial u of gv centre relative to blob origin
    bbor = bl(ii).bbor';
    gvb(ii).uc0bl = uc0im - bbor(:,ones(1,nv));
end

toc

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Apply strain distribution to grain
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Applying random deformation...')
tic

disp('Active deformation components:')
disp(fedpars.dcomps)

% Measured (true,real) strain distribution:
if ~exist('dmvol','var') || isempty(dmvol)
    [gvdm, dmvol, gradok] = gtFedTestGenerateRandomDef(fedpars, nd);
    gv.dm = gvdm;
else
    [gvdm, dmvol_size] = gtDefDmvol2Gvdm(dmvol);
    if (dmvol_size(4) == 3)
        mean_gvdm = mean(gvdm, 2);
        fprintf('Recentering GVDM and DMVOL to the mean orientation: (%f, %f, %f), from: (%f, %f, %f).\n', ...
            grain.R_vector, mean_gvdm)
        gvdm_shift = grain.R_vector' - mean_gvdm;
        gvdm = gvdm + gvdm_shift(:, ones(size(gvdm, 2), 1));
        dmvol = gtDefGvdm2Dmvol(gvdm, dmvol_size);
    end
    gv.dm = gvdm;
    gradok = [];
end


if ~all(gradok)
    disp('Gradients are higher than limit for components:')
    disp(find(~gradok)')
end

% figure('name','Epsilon33')
% slice(dmvol(:,:,:,3),grsize(1)/2,grsize(2)/2,grsize(3)/2);
% colorbar

disp('Largest deformation component:')
disp(max(dmvol(:)))
disp('Smallest deformation component:')
disp(min(dmvol(:)))

toc

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Simulate "measured" diffraction patterns
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

disp('Simulating true ("measured") blob intensities...')
gv.d = gv.dm;
[gvb, bl] = gtFedFwdProjExact(gv, gvb, bl, fedpars, parameters);

% Set back original zero actual strain
for ii = 1:nbl
    bl(ii).intm    = bl(ii).int;
    bl(ii).int(:)  = 0;
    bl(ii).intd(:) = bl(ii).int - bl(ii).intm;

    bl(ii).comim   = NaN(1,3);
    bl(ii).combl   = NaN(1,3);

    gvb(ii).ucbl8ini = gvb(ii).ucbl8;
end

disp('Calculate boundaries of measured blobs inside blob bounding box...')
for ii = 1:nbl
    uproj = sum(sum(abs(bl(ii).intm), 2), 3);
    bl(ii).mbbu = [find(uproj,1,'first'), find(uproj,1,'last')];

    vproj = sum(sum(abs(bl(ii).intm), 1), 3);
    bl(ii).mbbv = [find(vproj,1,'first'), find(vproj,1,'last')];

    wproj = sum(sum(abs(bl(ii).intm), 1), 2);
    bl(ii).mbbw = [find(wproj,1,'first'), find(wproj,1,'last')];

    bl(ii).mbbsize = [bl(ii).mbbu(2) - bl(ii).mbbu(1) + 1, ...
                      bl(ii).mbbv(2) - bl(ii).mbbv(1) + 1, ...
                      bl(ii).mbbw(2) - bl(ii).mbbw(1) + 1 ];
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Angle between consecutive blobs
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

for ii = 1:nbl
    nii = mod(ii, nbl) + 1;

    fedpars.blobnextang(ii) = abs(bl(ii).plsam*bl(nii).plsam');
    fedpars.blobnextang(ii) = min(fedpars.blobnextang(ii), 1);
    fedpars.blobnextang(ii) = acosd(fedpars.blobnextang(ii));

    fedpars.blobnextdet(ii) = abs(det(bl(ii).U0*bl(nii).U0'));
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Transform into single
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Transform into single precision
if strcmp(fedpars.gvtype, 'single')
    disp('Converting type to single...')
    fn = fieldnames(gv);
    for ii = 1:length(fn)
        gv.(fn{ii}) = single(gv.(fn{ii}));
    end

    fn = fieldnames(gvb);
    for ii = 1:length(fn)
        for jj = 1:length(gvb)
            gvb(jj).(fn{ii}) = single(gvb(jj).(fn{ii}));
        end
    end
end

disp('True blob dimensions:')
disp(vertcat(bl(:).mbbsize))

end % of main function
